@@ -1,4 +1,6 @@
 {-# LANGUAGE ScopedTypeVariables #-}
+{-# LANGUAGE ViewPatterns #-}
+{-# LANGUAGE MultilineStrings #-}
 {-# LANGUAGE BangPatterns #-}
 {-# LANGUAGE CPP #-}
 {-# LANGUAGE GeneralizedNewtypeDeriving #-}
@@ -11,17 +13,25 @@
 
 module GHC.Debugger.Monad where
 
+import System.Environment
+import System.Process
+import Control.Concurrent
+import Control.Concurrent.Async
+import Control.Exception
 import Control.Monad
 import Control.Monad.Catch
 import Control.Monad.IO.Class
 import Control.Monad.Reader
 import Data.Function
+import Data.Functor.Contravariant
 import Data.IORef
 import Data.Maybe
+import Data.Version (makeVersion, showVersion)
 import Prelude hiding (mod)
-import System.IO
+#ifdef MIN_VERSION_unix
 import System.Posix.Signals
-import qualified Data.IntMap as IM
+#endif
+import Data.Text (Text)
 import qualified Data.List as L
 import qualified Data.List.NonEmpty as NonEmpty
 
@@ -31,6 +41,8 @@ import GHC.Data.StringBuffer
 import GHC.Driver.Config.Diagnostic
 import GHC.Driver.DynFlags as GHC
 import GHC.Driver.Env
+import GHC.Driver.Monad
+import GHC.Driver.Hooks
 import GHC.Driver.Errors
 import GHC.Driver.Errors.Types
 import GHC.Driver.Main
@@ -46,19 +58,22 @@ import GHC.Types.SourceError
 import GHC.Types.SourceText
 import GHC.Types.Unique.Supply as GHC
 import GHC.Unit.Module.Graph
+import GHC.Unit.State
 import GHC.Unit.Module.ModSummary as GHC
 import GHC.Unit.Types
-import GHC.Utils.Logger as GHC
+import qualified GHC.Utils.Logger as GHC
 import GHC.Utils.Outputable as GHC
 import qualified GHC.LanguageExtensions as LangExt
 
 import GHC.Debugger.Interface.Messages
-import GHC.Debugger.Logger as Logger
-import GHC.Debugger.Runtime.Term.Cache
-import GHC.Debugger.Runtime.Term.Key
 import GHC.Debugger.Session
 import GHC.Debugger.Session.Builtin
+import GHC.Debugger.Runtime.Compile.Cache
+import GHC.Debugger.Utils
 import qualified GHC.Debugger.Breakpoint.Map as BM
+import qualified GHC.Debugger.Runtime.Thread.Map as TM
+
+import Colog.Core as Logger
 
 import {-# SOURCE #-} GHC.Debugger.Runtime.Instances.Discover (RuntimeInstancesCache, emptyRuntimeInstancesCache)
 
@@ -76,25 +91,14 @@ data DebuggerState = DebuggerState
         -- ^ Maps a 'InternalBreakpointId' in Trie representation (map of Module to map of Int) to the
         -- 'BreakpointStatus' it was activated with.
 
-      , varReferences     :: IORef (IM.IntMap TermKey, TermKeyMap Int)
-      -- ^ When we're stopped at a breakpoint, this maps variable reference to
-      -- Terms to allow further inspection and forcing by reference.
-      --
-      -- This map is only valid while stopped in this context. After stepping
-      -- or resuming evaluation in any available way, this map becomes invalid
-      -- and should therefore be cleaned.
-      --
-      -- The TermKeyMap map is a reverse lookup map to find which references
-      -- already exist for given names
-
-      , termCache         :: IORef TermCache
-      -- ^ TermCache
-
-      , rtinstancesCache :: IORef RuntimeInstancesCache
+      , rtinstancesCache  :: IORef RuntimeInstancesCache
       -- ^ RuntimeInstancesCache
 
-      , genUniq           :: IORef Int
-      -- ^ Generates unique ints
+      , threadMap         :: IORef TM.ThreadMap
+      -- ^ 'ThreadMap' for threads spawned by the debuggee
+
+      , compCache         :: IORef CompCache
+      -- ^ Cache loaded and compiled expressions.
 
       , hsDbgViewUnitId   :: Maybe UnitId
       -- ^ The unit-id of the companion @haskell-debugger-view@ unit, used for
@@ -115,9 +119,17 @@ data DebuggerState = DebuggerState
       --
       -- If the user explicitly disabled custom views, use @Nothing@.
 
-      , dbgLogger :: Recorder (WithSeverity DebuggerMonadLog)
+      , dbgLogger :: LogAction Debugger DebuggerLog
+      -- ^ See Note [Debugger, debuggee, and DAP logs]
       }
 
+instance GHC.HasLogger Debugger where
+  getLogger = liftGhc GHC.getLogger
+
+instance GHC.GhcMonad Debugger where
+  getSession = liftGhc GHC.getSession
+  setSession s = liftGhc $ GHC.setSession s
+
 -- | Enabling/Disabling a breakpoint
 data BreakpointStatus
       -- | Breakpoint is disabled
@@ -146,11 +158,14 @@ instance Outputable BreakpointStatus where ppr = text . show
 data RunDebuggerSettings = RunDebuggerSettings
       { supportsANSIStyling :: Bool
       , supportsANSIHyperlinks :: Bool
+      , preferInternalInterpreter :: Bool
+      , externalInterpreterStdinStream :: StdStream
+      -- ^ How to determine the stdin of the external interpreter running the
+      -- debuggee. If not using the external interpreter this field is unused.
       }
 
 -- | Run a 'Debugger' action on a session constructed from a given GHC invocation.
-runDebugger :: Recorder (WithSeverity DebuggerMonadLog)
-            -> Handle     -- ^ The handle to which GHC's output is logged. The debuggee output is not affected by this parameter.
+runDebugger :: LogAction IO DebuggerLog
             -> FilePath   -- ^ Cradle root directory
             -> FilePath   -- ^ Component root directory
             -> FilePath   -- ^ The libdir (given with -B as an arg)
@@ -161,11 +176,16 @@ runDebugger :: Recorder (WithSeverity DebuggerMonadLog)
             -> RunDebuggerSettings -- ^ Other debugger run settings
             -> Debugger a -- ^ 'Debugger' action to run on the session constructed from this invocation
             -> IO a
-runDebugger l dbg_out rootDir compDir libdir units ghcInvocation' extraGhcArgs mainFp conf (Debugger action) = do
+runDebugger l rootDir compDir libdir units ghcInvocation' extraGhcArgs mainFp conf (Debugger action) = do
+  let ghcLog = liftLogIO l :: LogAction Ghc DebuggerLog
+  let dbgLog = liftLogIO l :: LogAction Debugger DebuggerLog
+  thisProg <- getExecutablePath
   let ghcInvocation = filter (\case ('-':'B':_) -> False; _ -> True) ghcInvocation'
   GHC.runGhc (Just libdir) $ do
+#ifdef MIN_VERSION_unix
     -- Workaround #4162
     _ <- liftIO $ installHandler sigINT Default Nothing
+#endif
     dflags0 <- GHC.getSessionDynFlags
     let dflags1 = dflags0
           { GHC.ghcMode = GHC.CompManager
@@ -175,18 +195,36 @@ runDebugger l dbg_out rootDir compDir libdir units ghcInvocation' extraGhcArgs m
           , GHC.canUseErrorLinks = conf.supportsANSIHyperlinks
           }
           -- Default debugger settings
+          `GHC.xopt_set` LangExt.TypeApplications
           `GHC.xopt_set` LangExt.PackageImports
+          `GHC.xopt_set` LangExt.MagicHash -- needed for some of the expressions we compile
           `GHC.gopt_set` GHC.Opt_ImplicitImportQualified
           `GHC.gopt_set` GHC.Opt_IgnoreOptimChanges
           `GHC.gopt_set` GHC.Opt_IgnoreHpcChanges
           `GHC.gopt_set` GHC.Opt_UseBytecodeRatherThanObjects
           `GHC.gopt_set` GHC.Opt_InsertBreakpoints
+
+          -- Enable the external interpreter by default! See #169
+          -- See Note [Custom external interpreter]
+          & enableExternalInterpreter conf.preferInternalInterpreter
+          -- Ext interp is the same program as this, with "--external-interpreter"
+          -- (this is ignored on GHC 9.14, see Note [Custom external interpreter])
+          & setPgmI thisProg
+          -- ideally, we'd set "external-interpreter" *before* the file
+          -- descriptors. since there's no way to do that yet, we just have
+          -- some logic in main to detect [writefd, readfd, --external-interpreter]
+          & addOptI "--external-interpreter"
+
+          -- Really important to force -dynamic if host is dynamic
+          -- See Note [Dynamic Debuggee for dynamic debugger]
+          & enableDynamicDebuggee
+
           & setBytecodeBackend
           & enableByteCodeGeneration
 
     GHC.modifyLogger $
       -- Override the logger to output to the given handle
-      GHC.pushLogHook $ const $ debuggerLoggerAction dbg_out
+      GHC.pushLogHook $ const $ ghcLogAction l
 
     dflags2 <- getLogger >>= \logger -> do
       -- Set the extra GHC arguments for ALL units by setting them early in
@@ -199,13 +237,48 @@ runDebugger l dbg_out rootDir compDir libdir units ghcInvocation' extraGhcArgs m
       (dflags2, fileish_args, warns)
         <- parseDynamicFlags logger dflags1 (map noLoc extraGhcArgs)
       liftIO $ printOrThrowDiagnostics logger (initPrintConfig dflags2) (initDiagOpts dflags2) (GhcDriverMessage <$> warns)
-      -- todo: consider fileish_args?
       forM_ fileish_args $ \fish_arg -> liftIO $ do
-        logMsg logger MCOutput noSrcSpan $ text "Ignoring extraGhcArg which isn't a recognized flag:" <+> text (unLoc fish_arg)
+        GHC.logMsg logger MCOutput noSrcSpan $ text "Ignoring extraGhcArg which isn't a recognized flag:" <+> text (unLoc fish_arg)
         printOrThrowDiagnostics logger (initPrintConfig dflags2) (initDiagOpts dflags2) (GhcDriverMessage <$> warns)
       return dflags2
 
-    -- Set the session dynflags now to initialise the hsc_interp.
+    -- Make sure to override the function which creates the external
+    -- interpreter, because we need to keep track of the standard handles
+    iserv_handles <- liftIO newEmptyMVar
+    modifySession $ \h -> h
+      { hsc_hooks = (hsc_hooks h)
+          { createIservProcessHook = Just $ \cp -> do
+              -- See Note [External interpreter buffering]
+              (_, Just o, Just e, ph) <-
+                createProcess cp
+                  { std_in  = conf.externalInterpreterStdinStream
+                  , std_out = CreatePipe
+                  , std_err = CreatePipe
+                  -- Override executable path
+                  -- See Note [Custom external interpreter]
+#if MIN_VERSION_ghc(9,15,0)
+#else
+                  , cmdspec = case cmdspec cp of
+                      ShellCommand (words -> ws) -> ShellCommand $ unwords $ thisProg : drop 1 ws
+                      RawCommand _fp args -> RawCommand thisProg args
+#endif
+                  }
+              putMVar iserv_handles (o, e)
+              return ph
+          }
+      }
+
+    when (GHC.gopt GHC.Opt_ExternalInterpreter dflags2) $ liftIO $ void $ do
+      -- The external interpreter is spawned lazily, so we block waiting for
+      -- the handles to be available in a new thread.
+      forkIO $ do
+        withAsync (takeMVar iserv_handles) $ \async_handles -> do
+          (serv_out, serv_err) <- wait async_handles
+          concurrently_
+            (forwardHandleToLogger serv_err (contramap LogDebuggeeErr l))
+            (forwardHandleToLogger serv_out (contramap LogDebuggeeOut l))
+
+    -- Initializes interpreter!
     _ <- GHC.setSessionDynFlags dflags2
 
     -- Initialise plugins here because the plugin author might already expect this
@@ -217,10 +290,11 @@ runDebugger l dbg_out rootDir compDir libdir units ghcInvocation' extraGhcArgs m
 
     -- Discover the user-given flags and targets
     flagsAndTargets <- parseHomeUnitArguments mainFp compDir units ghcInvocation dflags2 rootDir
+    buildWays       <- liftIO $ validateUnitsWays flagsAndTargets
 
     -- Setup base HomeUnitGraph
     setupHomeUnitGraph (NonEmpty.toList flagsAndTargets)
-    -- Downsweep user-given modules first 
+    -- Downsweep user-given modules first
     mod_graph_base <- doDownsweep Nothing
 
     if_cache <- Just <$> liftIO newIfaceCache
@@ -234,13 +308,14 @@ runDebugger l dbg_out rootDir compDir libdir units ghcInvocation' extraGhcArgs m
 
         -- Add the custom unit to the HUG
         let base_dep_uids = [uid | UnitNode _ uid <- mg_mss mod_graph_base]
-        addInMemoryHsDebuggerViewUnit base_dep_uids =<< getDynFlags
+        addInMemoryHsDebuggerViewUnit base_dep_uids . setDynFlagWays buildWays =<< getDynFlags
 
         tryLoadHsDebuggerViewModule l if_cache (const False) debuggerViewClassModName debuggerViewClassContents
           >>= \case
             Failed -> do
               -- Failed to load base debugger-view module!
-              logWith l Debug $ LogFailedToCompileDebugViewModule debuggerViewClassModName
+              ghcLog <& DebuggerLog Logger.Debug
+                (LogFailedToCompileDebugViewModule debuggerViewClassModName)
               return []
             Succeeded -> (debuggerViewClassModName:) . concat <$> do
 
@@ -256,12 +331,14 @@ runDebugger l dbg_out rootDir compDir libdir units ghcInvocation' extraGhcArgs m
                           _ -> False) . GHC.targetId)
                       modName modContent >>= \case
                     Failed -> do
-                      logWith l Info $ LogFailedToCompileDebugViewModule modName
+                      ghcLog <& DebuggerLog Logger.Info
+                        (LogFailedToCompileDebugViewModule modName)
                       return []
                     Succeeded -> do
                       return [modName]
                 else do
-                  logWith l Debug $ LogSkippingViewModuleNoPkg modName pkgName (map unitIdString base_dep_uids)
+                  ghcLog <& DebuggerLog Logger.Debug
+                    (LogSkippingViewModuleNoPkg modName pkgName (map unitIdString base_dep_uids))
                   return []
 
       Just uid ->
@@ -307,8 +384,100 @@ runDebugger l dbg_out rootDir compDir libdir units ghcInvocation' extraGhcArgs m
         dbgViewImps ++
         map (GHC.IIModule . GHC.ms_mod) mss)
 
-    runReaderT action =<< initialDebuggerState l (if loadedBuiltinModNames == [] then Nothing else Just hdv_uid)
+    -- See Note [External interpreter buffering]
+    setBufferings <- compileExprRemote """
+      do { System.IO.hSetBuffering System.IO.stdout System.IO.LineBuffering
+         ; System.IO.hSetBuffering System.IO.stderr System.IO.LineBuffering }
+      """
+    hscInterp <$> GHC.getSession >>= \interp ->
+      liftIO $ evalIO interp setBufferings
 
+    runReaderT action
+      =<< initialDebuggerState dbgLog
+          (if loadedBuiltinModNames == []
+            then Nothing
+            else Just hdv_uid)
+
+{-
+Note [Custom external interpreter]
+~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
+We compile a custom external interpreter server with custom commands which make
+certain debugger operations possible in the remote process directly.
+This allows us to avoid excessive `Term` parsing and remote execution.
+(Note: we don't have custom commands just yet, but that is the vision)
+
+The custom external interpreter is the same executable as the debugger but invoked as:
+
+  hdb <write-fd> <read-fd> --external-interpreter
+
+(Note: external-interpreter is not the first argument because all `-opti`s are
+always inserted after the write-fd and read-fd.)
+
+When setting up the debugger session, we essentially set by default:
+  - Enable -fexternal-interpreter
+  - Set -pgmi=hdb and -opti=--external-interpreter
+This can be switched off by toggling `--internal-interpreter`
+
+With GHC 9.14, we have to override the `createProcess` executable call because
+of ghc's c94aaacd4c4 (GHC looks for a `-dyn` suffixed version of the custom
+external `-pgmi`, in this case `hdb` (but we do not have an `hdb-dyn`).
+In GHC 9.16 it is sufficient to specify the -pgmi.
+
+We can't use the on-the-fly external interpreter from GHC 9.14 because it is
+not compiled with -threaded (with 9.16 in principle could, but we really want
+the custom commands)
+
+Note: The custom external interpreter must be compiled with -fkeep-cafs!
+Why that is necessary is described in the GHC source code.
+
+Note [Dynamic Debuggee for dynamic debugger]
+~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
+A really really important point is that the debuggee MUST be linked dynamically
+if the debugger was compiled dynamically (checked with `hostIsDynamic`).
+
+Not doing this resulted in days upon days of suffering caused by a SIGILL fault.
+
+The bug surfaces when we load a non-PIC static object of the debuggee into a
+debugger that was linked dynamically with PIC. The runtime object linker does
+not handle this correctly and something goes very wrong with the relocated
+debuggee code.
+
+Notably, this bug didn't surface on macOS because the static objects are also
+compiled with -fPIC, and it didn't show up when using the distributed iserv
+executables because that has very few dyn-link-time dependencies and for some
+reason that doesn't trigger the bug. Adding more unused package dependencies
+was sufficient to re-trigger it on windows.
+
+Therefore, we always use -dynamic for compiling and loading the debuggee if the
+debugger is dynamic (`hostIsDynamic`).
+
+On Windows, the debugger will be static and we'll resort to statically linking
+the debuggee too. There won't be a PIC mismatch so this should work fine.
+
+Note [External interpreter buffering]
+~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
+When we launch the external interpreter process, we create pipes for
+stdin/stdout/stderr instead of inheriting the current process' handles.
+This allows us to cleanly separate the debuggee output from the debugger
+output, without needing to redirect handles or the like.
+
+However, using a pipe instead of a handle connected to a TTY means that, by
+default, the line buffering will be block based rather than line buffered.
+
+> Newly opened streams are normally fully buffered, with one exception: a
+  stream connected to an interactive device such as a terminal is initially
+  line buffered.^[1]
+
+We depend on line buffering to forward output from these handles to the
+debugger output (see `forwardHandleToLogger`).
+
+Therefore, after loading the modules, we evaluate on the remote process:
+
+  hSetBuffering stdout LineBuffering
+  hSetBuffering stderr LineBuffering
+
+[1] https://ftp.gnu.org/old-gnu/Manuals/glibc-2.2.5/html_node/Buffering-Concepts.html
+-}
 --------------------------------------------------------------------------------
 
 -- | Run downsweep on the currently set targets (see @hsc_targets@)
@@ -344,7 +513,7 @@ doLoad if_cache how_much mg = do
 -- | Returns @Just modName@ if the given module was successfully loaded
 tryLoadHsDebuggerViewModule
   :: GhcMonad m
-  => Recorder (WithSeverity DebuggerMonadLog)
+  => LogAction IO DebuggerLog
   -> Maybe ModIfaceCache
   -> (GHC.Target -> Bool)
   -- ^ Predicate to determine which of the existing
@@ -362,9 +531,9 @@ tryLoadHsDebuggerViewModule l if_cache keepTarget modName modContents = do
   -- modules we're trying to load and compile.
   restore_logger <- GHC.getLogger
   GHC.modifyLogger $
-    -- Emit it all as Debug-level logs
+    -- Emit it all as Debug-level debugger logs
     GHC.pushLogHook $ const $ \_ _ _ sdoc ->
-      logWith l Logger.Debug $ LogSDoc dflags sdoc
+      l <& DebuggerLog Logger.Debug (LogSDoc dflags sdoc)
 
   -- Make the target
   dvcT <- liftIO $ makeInMemoryHsDebuggerViewTarget modName modContents
@@ -381,7 +550,7 @@ tryLoadHsDebuggerViewModule l if_cache keepTarget modName modContents = do
 
   -- Restore logger
   GHC.modifyLogger $
-    GHC.pushLogHook (const $ putLogMsg restore_logger)
+    GHC.pushLogHook (const $ GHC.putLogMsg restore_logger)
 
 
   return result
@@ -403,6 +572,8 @@ getHsDebuggerViewUid = asks hsDbgViewUnitId
 -- See also comment on the @'hsDbgViewUnitId'@ field of @'DebuggerState'@
 findHsDebuggerViewUnitId :: ModuleGraph -> GHC.Ghc (Maybe UnitId)
 findHsDebuggerViewUnitId mod_graph = do
+  hsc_env <- getSession
+  let unitState = hsc_units hsc_env
 
   -- Only looks at unit-nodes, this is not robust!
   -- TODO: Better lookup of unit-id
@@ -410,79 +581,43 @@ findHsDebuggerViewUnitId mod_graph = do
         [ uid
         | UnitNode _deps uid <- mg_mss mod_graph
         , "haskell-debugger-view" `L.isPrefixOf` unitIdString uid
+            || "hskll-dbggr-vw" `L.isPrefixOf` unitIdString uid
+            || "haskell-debug_" `L.isPrefixOf` unitIdString uid
         ]
 
+      -- If the haskell-debugger-view is in the dependency graph, it must have
+      -- one of the versions the debugger is known to support:
+      supported_versions
+        = [ makeVersion [0, 2, 0, 0] ]
+
   case hskl_dbgr_vws of
-    [hdv_uid] ->
+    [hdv_uid] -> do
       -- In transitive closure, use that one.
-      return (Just hdv_uid)
+      -- Check that the version is exactly 0.2.0.0
+      case lookupUnit unitState (RealUnit (Definite hdv_uid)) of
+        Just unitInfo -> do
+          let version = unitPackageVersion unitInfo
+          if version `elem` supported_versions
+            then return (Just hdv_uid)
+            else throwM UnsupportedHsDbgViewVersion{supportedVersions=supported_versions, actualVersion=version}
+        Nothing ->
+          error "Could not find unit info for haskell-debugger-view"
     [] -> do
       return Nothing
     _  ->
       error "Multiple unit-ids found for haskell-debugger-view in the transitive closure?!"
 
---------------------------------------------------------------------------------
--- Variable references
---------------------------------------------------------------------------------
-
--- | Find a variable's associated Term and Name by reference ('Int')
-lookupVarByReference :: Int -> Debugger (Maybe TermKey)
-lookupVarByReference i = do
-  ioref <- asks varReferences
-  (rm, _) <- readIORef ioref & liftIO
-  return $ IM.lookup i rm
-
--- | Finds or creates an integer var reference for the given 'TermKey'.
--- TODO: Arguably, this mapping should be part of the debug-adapter, and
--- haskell-debugger should deal in 'TermKey' terms only.
-getVarReference :: TermKey -> Debugger Int
-getVarReference key = do
-  ioref     <- asks varReferences
-  (rm, tkm) <- readIORef ioref & liftIO
-  (i, tkm') <- case lookupTermKeyMap key tkm of
-    Nothing -> do
-      new_i <- freshInt
-      return (new_i, insertTermKeyMap key new_i tkm)
-    Just existing_i ->
-      return (existing_i, tkm)
-  let rm' = IM.insert i key rm
-  writeIORef ioref (rm', tkm') & liftIO
-  return i
-
--- | Whenever we run a request that continues execution from the current
--- suspended state, such as Next,Step,Continue, this function should be called
--- to delete the variable references that become invalid as we leave the
--- suspended state.
---
--- In particular, @'varReferences'@ is reset.
---
--- See also section "Lifetime of Objects References" in the DAP specification.
-leaveSuspendedState :: Debugger ()
-leaveSuspendedState = do
-  ioref <- asks varReferences
-  liftIO $ writeIORef ioref mempty
-
 --------------------------------------------------------------------------------
 -- Utilities
 --------------------------------------------------------------------------------
 
--- | Generate a new unique 'Int'
-freshInt :: Debugger Int
-freshInt = do
-  ioref <- asks genUniq
-  i <- readIORef ioref & liftIO
-  let !i' = i+1
-  writeIORef ioref i'  & liftIO
-  return i
-
 -- | Initialize a 'DebuggerState'
-initialDebuggerState :: Recorder (WithSeverity DebuggerMonadLog) -> Maybe UnitId -> GHC.Ghc DebuggerState
+initialDebuggerState :: LogAction Debugger DebuggerLog -> Maybe UnitId -> GHC.Ghc DebuggerState
 initialDebuggerState l hsDbgViewUid =
   DebuggerState <$> liftIO (newIORef BM.empty)
-                <*> liftIO (newIORef mempty)
-                <*> liftIO (newIORef mempty)
                 <*> liftIO (newIORef emptyRuntimeInstancesCache)
-                <*> liftIO (newIORef 0)
+                <*> liftIO (newIORef TM.emptyThreadMap)
+                <*> liftIO (newIORef emptyCompCache)
                 <*> pure hsDbgViewUid
                 <*> pure l
 
@@ -495,7 +630,23 @@ instance Exception DebuggerFailedToLoad
 instance Show DebuggerFailedToLoad where
   show DebuggerFailedToLoad = "Failed to compile and load user project."
 
---------------------------------------------------------------------------------
+data UnsupportedHsDbgViewVersion = UnsupportedHsDbgViewVersion
+  { supportedVersions :: [ Version ]
+  , actualVersion :: Version
+  }
+instance Exception UnsupportedHsDbgViewVersion
+instance Show UnsupportedHsDbgViewVersion where
+  show (UnsupportedHsDbgViewVersion supported actual) =
+    "Cannot use unsupported haskell-debugger-view version found in the transitive closure: " ++ showVersion actual ++
+    " (supported: " ++ L.intercalate ", " (map showVersion supported) ++ ")"
+
+expectRight :: Exception e => Either e a -> Debugger a
+expectRight s = case s of
+  Left e -> do
+    logSDoc Logger.Error (text $ displayException e)
+    liftIO $ throwIO e
+  Right a -> do
+    pure a
 
 --------------------------------------------------------------------------------
 -- * Modules
@@ -550,52 +701,41 @@ deepseqTerm hsc_env t = case t of
                        return t{wrapped_term = wrapped_term'}
   _              -> do seqTerm hsc_env t
 
---------------------------------------------------------------------------------
--- * Instances
---------------------------------------------------------------------------------
-
-instance GHC.HasLogger Debugger where
-  getLogger = liftGhc GHC.getLogger
-
-instance GHC.GhcMonad Debugger where
-  getSession = liftGhc GHC.getSession
-  setSession s = liftGhc $ GHC.setSession s
-
 --------------------------------------------------------------------------------
 -- * Logging
 --------------------------------------------------------------------------------
 
--- | The logger action used to log GHC output
-debuggerLoggerAction :: Handle -> GHC.LogAction
-debuggerLoggerAction h a b c d = do
-  hSetEncoding h utf8 -- GHC output uses utf8
-  -- potentially use the `Recorder` here?
-  defaultLogActionWithHandles h h a b c d
+-- | A debugger log. May include debuggee ouput.
+data DebuggerLog
+  = DebuggerLog !Logger.Severity !DebuggerMessage
+  | GHCLog !GHC.LogFlags !MessageClass !SrcSpan !SDoc
+  | LogDebuggeeOut !Text
+  | LogDebuggeeErr !Text
 
-data DebuggerMonadLog
-  = LogFailedToCompileDebugViewModule GHC.ModuleName
-  | LogSkippingViewModuleNoPkg GHC.ModuleName String [String]
-  | LogSDoc DynFlags SDoc
+-- | A debugger log message
+data DebuggerMessage
+  = LogSDoc !DynFlags !SDoc
+  | LogFailedToCompileDebugViewModule !GHC.ModuleName
+  | LogSkippingViewModuleNoPkg !GHC.ModuleName String [String]
 
-instance Pretty DebuggerMonadLog where
-  pretty = \ case
+instance Show DebuggerMessage where
+  show = \ case
     LogFailedToCompileDebugViewModule mn ->
-      pretty $ "Failed to compile built-in " ++ moduleNameString mn ++ " module! Ignoring these custom debug views."
+      "Failed to compile built-in " ++ moduleNameString mn ++ " module! Ignoring these custom debug views."
     LogSkippingViewModuleNoPkg mn pkg uids ->
-      pretty $ "Skipping compilation of built-in " ++ moduleNameString mn ++ " module because package "
-                ++ show pkg ++ " wasn't found in dependencies " ++ show uids
-    LogSDoc dflags doc ->
-      pretty $ showSDoc dflags doc
+      "Skipping compilation of built-in " ++ moduleNameString mn ++ " module because package "
+          ++ show pkg ++ " wasn't found in dependencies " ++ show uids
+    LogSDoc dflags doc -> showSDoc dflags doc
 
 logSDoc :: Logger.Severity -> SDoc -> Debugger ()
 logSDoc sev doc = do
   dflags <- getDynFlags
   l <- asks dbgLogger
-  logWith l sev (LogSDoc dflags doc)
+  l <& DebuggerLog sev (LogSDoc dflags doc)
 
-logAction :: Recorder (WithSeverity DebuggerMonadLog) -> DynFlags -> GHC.LogAction
-logAction l dflags = \_ msg_class _ sdoc -> do
-    logWith l (msgClassSeverity msg_class) $ LogSDoc dflags sdoc
+ghcLogAction :: LogAction IO DebuggerLog -> GHC.LogAction
+ghcLogAction l = \logflags mclass srcSpan sdoc -> do
+    liftLogIO l <& GHCLog logflags mclass srcSpan sdoc
 
 msgClassSeverity :: MessageClass -> Logger.Severity
 msgClassSeverity = \case
@@ -604,10 +744,6 @@ msgClassSeverity = \case
   MCInteractive -> Info
   MCDump -> Debug
   MCInfo -> Info
-  MCDiagnostic sev _ _ -> ghcSevSeverity sev
-
-ghcSevSeverity :: GHC.Severity -> Logger.Severity
-ghcSevSeverity = \case
-  SevIgnore -> Debug -- ?
-  SevWarning -> Logger.Warning
-  SevError -> Logger.Error
+  MCDiagnostic SevIgnore _ _ -> Debug -- ?
+  MCDiagnostic SevWarning _ _ -> Logger.Warning
+  MCDiagnostic SevError _ _ -> Logger.Error
