@@ -7,6 +7,7 @@
 module GHC.Debugger.Session (
   parseHomeUnitArguments,
   setupHomeUnitGraph,
+  validateUnitsWays,
   TargetDetails(..),
   Target(..),
   toGhcTarget,
@@ -21,6 +22,10 @@ module GHC.Debugger.Session (
   setCacheDirs,
   setBytecodeBackend,
   enableByteCodeGeneration,
+  enableExternalInterpreter,
+  enableDynamicDebuggee,
+  setPgmI, addOptI,
+  setDynFlagWays
   )
   where
 
@@ -41,6 +46,7 @@ import qualified System.Directory as Directory
 import qualified System.Environment as Env
 
 import qualified GHC
+import GHC.Platform.Ways
 import GHC.Driver.DynFlags as GHC
 import GHC.Driver.Monad
 import qualified GHC.Driver.Session as GHC
@@ -52,9 +58,11 @@ import GHC.Unit.Types
 import qualified GHC.Unit.State                        as State
 import GHC.Driver.Env
 import GHC.Types.SrcLoc
+import GHC.Settings (ToolSettings(..))
 import Language.Haskell.Syntax.Module.Name
 import qualified Data.Foldable as Foldable
 import qualified GHC.Unit.Home.Graph as HUG
+import qualified Data.Set as Set
 import Data.Maybe
 import GHC.Types.Target (InputFileBuffer)
 
@@ -163,8 +171,7 @@ initHomeUnitEnv unitDflags env = do
   initial_home_graph <- createUnitEnvFromFlags dflags0 unitDflags
   let home_units = unitEnv_keys initial_home_graph
   init_home_unit_graph <- forM initial_home_graph $ \homeUnitEnv -> do
-    let cached_unit_dbs = homeUnitEnv_unit_dbs homeUnitEnv
-        dflags = homeUnitEnv_dflags homeUnitEnv
+    let dflags = homeUnitEnv_dflags homeUnitEnv
         old_hpt = homeUnitEnv_hpt homeUnitEnv
     (dbs,unit_state,home_unit,mconstants) <- State.initUnits (hsc_logger env) dflags Nothing home_units
     updated_dflags <- GHC.updatePlatformConstants dflags mconstants
@@ -230,6 +237,17 @@ setupMultiHomeUnitGhcSession exts hsc_env cis = do
       return (L.nubOrdOn targetTarget ctargets)
     pure (hscEnv', concat ts)
 
+-- | Find and return the ways in which the home units are built.
+-- INVARIANT: All home units are built with the same 'Ways'
+validateUnitsWays :: NonEmpty.NonEmpty (DynFlags, [GHC.Target]) -> IO Ways
+validateUnitsWays flagsAndTargets = do
+    let unitWays  = NonEmpty.map (ways . fst) flagsAndTargets
+        firstWays = NonEmpty.head unitWays
+        restWays  = NonEmpty.tail unitWays
+    if all (== firstWays) restWays
+        then return firstWays
+        else error "Unexpected: the home units have different ways! Not supported, see GHC#26765"
+
 data TargetDetails = TargetDetails
   { targetTarget :: Target
   -- ^ Simplified version of 'TargetId', storing enough information
@@ -406,6 +424,47 @@ setBytecodeBackend dflags = dflags
 #endif
   }
 
+-- | Enable the external interpreter by default unless the user sets
+-- @preferInternalInterpreter=True@ (with @--internal-interpreter@)
+enableExternalInterpreter :: Bool -> DynFlags -> DynFlags
+enableExternalInterpreter preferInternalInterpreter dflags
+  | preferInternalInterpreter
+  = dflags `GHC.gopt_unset` GHC.Opt_ExternalInterpreter
+  | otherwise
+  = dflags `GHC.gopt_set` GHC.Opt_ExternalInterpreter
+
+-- | Force -dynamic on the debuggee if the debugger (which is also the external
+-- interpreter) was compiled with -dynamic. On Windows the debugger can't be
+-- built dynamic, so we won't enable it there.
+--
+-- See Note [Dynamic Debuggee for dynamic debugger]
+enableDynamicDebuggee :: DynFlags -> DynFlags
+enableDynamicDebuggee dflags
+  | hostIsDynamic
+  = addWay' WayDyn dflags
+  | otherwise
+  = dflags
+
+setPgmI, addOptI :: String -> DynFlags -> DynFlags
+setPgmI f = alterToolSettings $ \s -> s { toolSettings_pgm_i = f }
+addOptI f = alterToolSettings $ \s -> s { toolSettings_opt_i = f : toolSettings_opt_i s }
+
+alterToolSettings :: (ToolSettings -> ToolSettings) -> DynFlags -> DynFlags
+alterToolSettings f dynFlags = dynFlags { toolSettings = f (toolSettings dynFlags) }
+
+setDynFlagWays :: Ways -> DynFlags -> DynFlags
+setDynFlagWays ws dyn = Set.foldr addWay' dyn ws
+
+addWay' :: Way -> DynFlags -> DynFlags
+addWay' w dflags0 =
+   let platform = targetPlatform dflags0
+       dflags1 = dflags0 { targetWays_ = addWay w (targetWays_ dflags0) }
+       dflags2 = foldr GHC.setGeneralFlag' dflags1
+                       (wayGeneralFlags platform w)
+       dflags3 = foldr GHC.unSetGeneralFlag' dflags2
+                       (wayUnsetGeneralFlags platform w)
+   in dflags3
+
 -- ----------------------------------------------------------------------------
 -- Utils that we need, but don't want to incur an additional dependency for.
 -- ----------------------------------------------------------------------------
