@@ -21,11 +21,11 @@ import Development.Debug.Adapter.Interface
 commandBreakpointLocations :: DebugAdaptor ()
 commandBreakpointLocations = do
   BreakpointLocationsArguments{..} <- getArguments
-  file <- fileFromSourcePath breakpointLocationsArgumentsSource
+  filePath <- fileFromSourcePath breakpointLocationsArgumentsSource
 
   DidGetBreakpoints mspan <-
     sendSync $ GetBreakpointsAt
-      ModuleBreak { path      = file
+      ModuleBreak { path      = filePath
                   , lineNum   = breakpointLocationsArgumentsLine
                   , columnNum = breakpointLocationsArgumentsColumn
                   }
@@ -47,17 +47,17 @@ commandBreakpointLocations = do
 commandSetBreakpoints :: DebugAdaptor ()
 commandSetBreakpoints = do
   SetBreakpointsArguments {..} <- getArguments
-  file <- fileFromSourcePath setBreakpointsArgumentsSource
+  filePath <- fileFromSourcePath setBreakpointsArgumentsSource
   let breaks_wanted = fromMaybe [] setBreakpointsArgumentsBreakpoints
 
   -- Clear existing module breakpoints
-  DidClearBreakpoints <- sendSync (ClearModBreakpoints file)
+  DidClearBreakpoints <- sendSync (ClearModBreakpoints filePath)
 
   -- Set requested ones
   breaks <- forM breaks_wanted $ \bp -> do
     DidSetBreakpoint bf <-
       sendSync $ SetBreakpoint
-        ModuleBreak { path      = file
+        ModuleBreak { path      = filePath
                     , lineNum   = DAP.sourceBreakpointLine bp
                     , columnNum = DAP.sourceBreakpointColumn bp
                     }
@@ -81,7 +81,7 @@ commandSetFunctionBreakpoints = do
   breaks <- forM breaks_wanted $ \bp -> do
     DidSetBreakpoint bf <-
       sendSync $ SetBreakpoint
-        FunctionBreak { function  = T.unpack $ fromJust {- not optional! -} $ DAP.functionBreakpointName bp }
+        FunctionBreak { function  = T.unpack $ DAP.functionBreakpointName bp }
         (readMaybe @Int =<< (T.unpack <$> DAP.functionBreakpointHitCondition bp))
         (T.unpack <$> DAP.functionBreakpointCondition bp)
     registerBreakFound bf
@@ -141,20 +141,13 @@ registerBreakFound b =
         , DAP.breakpointId = Just bid
         }) bids
 
--- | Adds new BreakpointId for a givent StgPoint
+-- | Adds new BreakpointId to the debug adapter mapping
 registerNewBreakpoint :: GHC.InternalBreakpointId -> DebugAdaptor BreakpointId
 registerNewBreakpoint breakpoint = do
-  bkpId <- getFreshBreakpointId
+  bkpId <- getFreshId
   updateDebugSession $ \das@DAS{..} -> das {breakpointMap = Map.insertWith mappend breakpoint (IS.singleton bkpId) breakpointMap}
   pure bkpId
 
--- | Generate fresh breakpoint Id.
-getFreshBreakpointId :: DebugAdaptor BreakpointId
-getFreshBreakpointId = do
-  bkpId <- nextFreshBreakpointId <$> getDebugSession
-  updateDebugSession $ \s -> s { nextFreshBreakpointId = nextFreshBreakpointId s + 1 }
-  pure bkpId
-
 -- | Get the file from a DAP Source
 --
 -- TODO: Handles sourceReferences too
