@@ -10,7 +10,7 @@ module Development.Debug.Session.Setup
   , hieBiosSetup
 
   -- * Logging
-  , FlagsLog(..)
+  , SessionSetupLog(..)
   ) where
 
 import Control.Applicative ((<|>))
@@ -29,7 +29,7 @@ import System.Directory hiding (findFile)
 import System.FilePath
 import System.IO.Error
 import Text.ParserCombinators.ReadP (readP_to_S)
-import Prettyprinter
+import Data.Functor.Contravariant
 
 import qualified Data.Text as T
 
@@ -42,18 +42,13 @@ import qualified Hie.Cabal.Parser as Implicit
 import qualified Hie.Locate as Implicit
 import qualified Hie.Yaml as Implicit
 
-import GHC.Debugger.Logger
+import Colog.Core
 
-data FlagsLog
+data SessionSetupLog
   = HieBiosLog HIE.Log
   | LogCradle (HIE.Cradle Void)
   | LogSetupMsg T.Text
-
-instance Pretty FlagsLog where
-  pretty = \ case
-    HieBiosLog msg -> pretty msg
-    LogCradle crdl -> "Determined Cradle:" <+> viaShow crdl
-    LogSetupMsg txt -> pretty txt
+  deriving Show
 
 -- | Flags inferred by @hie-bios@ to invoke GHC
 data HieBiosFlags = HieBiosFlags
@@ -69,49 +64,48 @@ data HieBiosFlags = HieBiosFlags
       }
 
 -- | Prepare a GHC session using hie-bios from scratch
-hieBiosSetup :: Recorder (WithSeverity FlagsLog)
+hieBiosSetup :: LogAction IO (WithSeverity SessionSetupLog)
              -> FilePath -- ^ project root
              -> FilePath -- ^ entry file
              -> ExceptT String IO (Either String HieBiosFlags)
 hieBiosSetup logger projectRoot entryFile = do
 
-  logT "Figuring out the right flags to compile the project using hie-bios..."
+  logInfo "Figuring out the right flags to compile the project using hie-bios..."
   cradle <- hieBiosCradle logger projectRoot entryFile & ExceptT
 
   -- GHC is found in PATH (by hie-bios as well).
-  logT "Checking GHC version against debugger version..."
-  _version <- hieBiosRuntimeGhcVersion logger cradle
+  logInfo "Checking GHC version against debugger version..."
+  _version <- hieBiosRuntimeGhcVersion cradle
 
-  logT "Discovering session flags with hie-bios..."
-  r <- hieBiosFlags logger cradle projectRoot entryFile     & liftIO
+  logInfo "Discovering session flags with hie-bios..."
+  r <- hieBiosFlags cradle projectRoot entryFile     & liftIO
 
-  logT "Session setup with hie-bios was successful."
+  logInfo "Session setup with hie-bios was successful."
   return r
 
   where
-    logT = logWith logger Info . LogSetupMsg . T.pack
+    logInfo m = liftLogIO logger <& WithSeverity (LogSetupMsg (T.pack m)) Info
 
 -- | Try implicit-hie and the builtin search to come up with a @'HIE.Cradle'@
-hieBiosCradle :: Recorder (WithSeverity FlagsLog) {-^ Logger -}
-              -> FilePath {-^ Project root -}
-              -> FilePath {-^ Entry file relative to root -}
+hieBiosCradle :: LogAction IO (WithSeverity SessionSetupLog)
+              -> FilePath -- ^ Project root
+              -> FilePath -- ^ Entry file relative to root
               -> IO (Either String (HIE.Cradle Void))
 hieBiosCradle logger root relTarget = runExceptT $ do
   let target = root </> relTarget
   explicitCradle <- HIE.findCradle target & liftIO
   cradle <- maybe (loadImplicitCradle hieBiosLogger target)
                   (HIE.loadCradle hieBiosLogger) explicitCradle & liftIO
-  logWith logger Info $ LogCradle cradle
+  liftLogIO logger <& WithSeverity (LogCradle cradle) Info
   pure cradle
   where
-    hieBiosLogger = toCologAction $ cmapWithSev HieBiosLog logger
+    hieBiosLogger = contramap (fmap HieBiosLog) logger
 
 -- | Fetch the runtime GHC version, according to hie-bios, and check it is the
 -- same as the compile time GHC version
-hieBiosRuntimeGhcVersion :: Recorder (WithSeverity FlagsLog)
-                         -> HIE.Cradle Void
+hieBiosRuntimeGhcVersion :: HIE.Cradle Void
                          -> ExceptT String IO Version
-hieBiosRuntimeGhcVersion _logger cradle = do
+hieBiosRuntimeGhcVersion cradle = do
   out <- liftIO (HIE.getRuntimeGhcVersion cradle) >>= unwrapCradleResult "Failed to get runtime GHC version"
 
   case versionMaybe out of
@@ -129,12 +123,11 @@ hieBiosRuntimeGhcVersion _logger cradle = do
       pure actualVersion
 
 -- | Make 'HieBiosFlags' from the given target file
-hieBiosFlags :: Recorder (WithSeverity FlagsLog) {-^ Logger -}
-             -> HIE.Cradle Void {-^ Project cradle the entry file belongs to -}
+hieBiosFlags :: HIE.Cradle Void {-^ Project cradle the entry file belongs to -}
              -> FilePath {-^ Project root -}
              -> FilePath {-^ Entry file relative to root -}
              -> IO (Either String HieBiosFlags)
-hieBiosFlags _logger cradle root relTarget = runExceptT $ do
+hieBiosFlags cradle root relTarget = runExceptT $ do
   let target = root </> relTarget
   libdir <- liftIO (HIE.getRuntimeGhcLibDir cradle) >>= unwrapCradleResult "Failed to get runtime GHC libdir"
 
